/* tund.c - tunnel daemon */

/*
 * Copyright (c) 1988, Julian Onions. 
 *
 * This source may be freely distributed, however I would be interested
 * in any changes that are made.
 */

/*
 * $Header: tund.c,v 1.10 88/07/11 08:31:38 jpo Exp $
 *
 * $Log:	tund.c,v $
 * Revision 1.10  88/07/11  08:31:38  jpo
 * Some clean ups.
 * 
 * Revision 1.10  88/07/11  08:29:04  jpo
 * Some tidying up
 * 
 * Revision 1.9  88/06/30  14:42:20  jpo
 * Ignore protocol ID on match
 * 
 * Revision 1.8  88/06/29  16:29:50  jpo
 * New CUDF handling
 * 
 * Revision 1.7  88/01/25  14:20:43  jpo
 * Arrgghhh brain damage!!
 * 
 * Revision 1.6  88/01/25  14:00:23  jpo
 * Added the mask stuff - doesn't work with sunlink tcp/ip stuff
 * 
 * Revision 1.5  87/12/10  09:12:26  jpo
 * Minor fixup
 * 
 * Revision 1.4  87/12/02  19:42:22  jpo
 * Bug: doesn't keep SA components, does now.
 * 
 * Revision 1.3  87/11/04  14:29:36  jpo
 * Added daemon startup mode - default.
 * 
 * Revision 1.2  87/10/28  17:59:39  jpo
 * More debugging etc + facilities.
 * 
 * Revision 1.1  87/10/20  09:41:46  jpo
 * Initial revision
 * 
 */

#ifndef lint
static char Rcsid[] = "$Header: tund.c,v 1.10 88/07/11 08:31:38 jpo Exp $";
#endif

#include "tund.h"

static void adios (), oob_data();
void    advise ();
char	*strdup (), *getstr (), *px25addr (), *rindex (), *ptime ();
extern	char	*strncpy (), *sprintf ();
extern  time_t time ();

typedef struct connection {
	char	*device;
	char	*interface;
	char	*if_addr;
	char	*dst_addr;
	int	flags;
#define	CON_MASTER		001
#define CON_SLAVE		002
#define CON_OOD			004
	time_t	contime;
	time_t	lastuse;
	int	tunfd, x25fd;
	CONN_DB	x25remote;
	CONN_DB x25local;
} Connection;

Connection	cons[MAXCONS];
Connection	*fd2con[32];

CONN_DB	listeners[MAXCONS];
CONN_DB	myx25addr;

FACILITY_DB facilities = {
	0,	/* reverse charge */
	0,	/* recv packetsize */
	0,	/* send packet size */
	0,	/* recv window size */
	0,	/* send window size */
	0,	/* recv through put */
	0,	/* send through put */
	0,	/* cug_req */
	0,	/* cug_index */
	0,	/* fast select type */
	0,	/* rpoa request */
	0	/* rpoa */
};

char	*myname;
char	*usage = "Usage: %s [-d] configfile";

int	debug;
int	usemasks;
int	readfds;
int	listenfds = 0;
int     x25fds = 0;
int	maxlistening = 0;
int	timeout = TIMEOUT;
int	x25open = 0;
time_t	currenttime;

main (argc, argv)
char	      **argv;
{
	extern char    *optarg;
	extern int	optind;
	int	opt;

	if (myname = rindex(argv[0], '/'))
		myname ++;
	else	myname = argv[0];

	while ((opt = getopt (argc, argv, "mdt:")) != EOF)
		switch (opt)
		{
		    case 'd':
			debug = TRUE;
			setlinebuf(stdout);
			break;
		    case 't':
			timeout = atoi(optarg);
			if (timeout <= 0)
				adios (NULLCP, "Timeout must be greater than 0");
			break;
		    case 'm':
			usemasks = 1;
			break;
		}
	argc -= optind;
	argv += optind;

	if (argc <= 0 )
		adios (NULLCP, usage, myname);
	read_data (*argv);

	envinit ();
	init_devices ();
	(void) signal (SIGPIPE, SIG_IGN);
	(void) signal (SIGURG, oob_data);
	doit ();
	exit (1);
}

init_devices ()
{
	Connection *cp;

	for (cp = cons; cp -> device; cp ++)
	{
		if (cp -> flags & CON_MASTER) {
			if (initdev(cp) == NOTOK)
				exit (1);
		}
		if (cp -> flags & CON_SLAVE)
			initlisten (cp);
	}
}

initdev (cp)
Connection *cp;
{
	if ((cp -> tunfd = open (cp -> device, 2)) == NOTOK) {
		advise (cp -> device, LOG_WARNING, "Can't open");
		return NOTOK;
	}
	setaddresses (cp);
	readfds |= (1 << cp -> tunfd);
	fd2con[cp -> tunfd] = cp;
	return 0;
}

initlisten (cp)
Connection *cp;
{
	int	sd, i;

	for (i = 0; i < maxlistening; i++)
		 if (x25cmp(&cp->x25local, &listeners[i])) {
#ifdef DEBUG
			advise (NULLCP, LOG_DEBUG, "Already listening on %s",
				px25addr(&cp->x25local));
#endif
			return;
		}

	listeners[maxlistening++] = cp -> x25local;
	if((sd = start_x25_server (&cp->x25local, 5, 0, 0)) == NOTOK)
		adios ("failed", "start_x25_server");
	readfds |= (1 << sd);
	listenfds |= (1 << sd);
	fd2con[sd] = cp;
	x25fds |= (1 << sd);
}

setaddresses (cp)
Connection *cp;
{
	struct	sockaddr_in	*sin;
	int	fd;
	struct  ifreq   ifr;
	
#ifdef DEBUG
	printcp ("set addresses", cp);
#endif
	bzero ((char *)&ifr, sizeof ifr);
	(void) strncpy (ifr.ifr_name, cp -> interface, sizeof ifr.ifr_name);
	if ((fd = socket (AF_INET, SOCK_DGRAM, 0)) < 0)
		adios ("socket", "Can't open");
	sin = (struct sockaddr_in *) &(ifr.ifr_addr);
	getaddr (cp -> if_addr, sin);
	if (ioctl (fd, SIOCSIFADDR, (caddr_t) &ifr) < 0)
		adios ("SIOCSIFADDR", "ioctl");
	sin = (struct sockaddr_in *)&ifr.ifr_dstaddr;
	getaddr (cp -> dst_addr, sin);
	if (ioctl (fd, SIOCSIFDSTADDR, (caddr_t) &ifr) < 0)
		adios ("SIOCSIFDSTADDR", "ioctl");
	(void) close (fd);
}

doit ()
{
	int	rfds, i;
	int	nfds = 32;
	struct timeval timer;
	struct timeval *tv = &timer;

	tv -> tv_usec = 500000;
	tv -> tv_sec = timeout;
#define the_universe_exists	1
	/* stamp out boring names committee strikes again */
	
	while (the_universe_exists)
	{
		rfds = readfds;
		tv = x25open > 0 ? &timer : (struct timeval *)0;
#ifdef DEBUG
		advise (NULLCP, LOG_DEBUG,
			"select on 0x%x x25vc=%d timeout=%d",
			readfds, x25open, tv ? tv->tv_sec : -1);
#endif
		switch (select (nfds, &rfds, NULLIP, NULLIP, tv))
		{
		    case NOTOK:
			advise("failed", LOG_WARNING,"select");
			continue;

		    case 0:
			(void) time(&currenttime);
			close_unused ();
			continue;
		}
		(void) time(&currenttime);
		for (i = 0; i < nfds; i++)
			if (rfds & (1 << i)) {
				if (listenfds & (1 << i))
					list_conn (i);
				else	read_conn (i);
			}
		close_unused ();
	}
}

list_conn (fd)
int	fd;
{
	CONN_DB	zremote;
	CONN_DB *remote = &zremote;
	int	x25fd;
	Connection *cp;

	if((x25fd = join_x25_client (fd, remote)) == NOTOK)
	{
		advise ("failed", LOG_WARNING, "join_x25_client");
		return;
	}
	for (cp = cons; cp -> device; cp++) {
		if ((cp -> flags & CON_SLAVE) == 0)
			continue;
		if ( cp -> x25fd != NOTOK)
			continue;
		if (x25cmp(remote, &cp->x25remote))
			break;
	}
	if (cp -> device == NULLCP) {
		advise (NULLCP, LOG_WARNING, "No match for incoming call %s",
			px25addr(remote));
		(void) x25_close (x25fd);
		return;
	}
	if (cp -> tunfd == NOTOK) {
		if (initdev (cp) == NOTOK)
			return;
	}
	set_x25opts(x25fd);
	cp -> x25fd = x25fd;
	cp -> contime = cp -> lastuse = currenttime;

	readfds |= (1 << cp-> x25fd);
	x25fds |= (1 << cp-> x25fd);
	fd2con[cp -> x25fd] = cp;
	advise (NULLCP, LOG_INFO, "new connection accepted from %s",
		px25addr(remote));
}

void read_conn (fd)
int	fd;
{
	char	buffer[TUNMTU];
	Connection *cp = fd2con[fd];
	int	n, writ;
	extern int errno;

#ifdef DEBUG
	advise (NULLCP, LOG_DEBUG, "read on %d", fd);
	printcp("read conn", cp);
#endif
	n = read(fd, buffer, sizeof(buffer));
	switch (n)
	{
	    case NOTOK:
		advise ("error", LOG_WARNING, "read");
		return;
	    case 0:
#ifdef DEBUG
		advise (NULLCP, LOG_DEBUG, "CLosing connection");
#endif
		close_conn (cp, fd, "connection dropped");
		return;
	}
	cp -> lastuse = currenttime;
	if (fd == cp -> tunfd ) {
		if (cp -> x25fd == NOTOK) {
			if (open_conn (cp) == NOTOK)
				return;
		}
		if ((writ = write (cp -> x25fd, buffer, n)) != n) {
			advise ("failed", LOG_WARNING, "write %d/%d", n, writ);
			close_conn (cp, cp -> x25fd, "write error");
		}
#ifdef DEBUG
		advise (NULLCP, LOG_DEBUG, "Written %d bytes to X25", writ);
#endif
	}
	else if (fd == cp -> x25fd ) {
		if ((writ = write (cp -> tunfd, buffer, n)) != n) {
			advise("failed", "DROPPED: write to tunnel %d/%d", n, writ);
		}
#ifdef DEBUG
		advise (NULLCP, LOG_DEBUG, "Written %d bytes to Tunnel", writ);
#endif
	}
	else adios (NULLCP, "Connection table does not match file descriptor");
}

getaddr (s, sin)
char	       *s;
struct sockaddr_in *sin;
{
	struct hostent *hp;
	struct netent  *np;
	long		val;

	hp = gethostbyname (s);
	if (hp)
	{
		sin->sin_family = hp->h_addrtype;
		bcopy (hp->h_addr, (char *) &sin->sin_addr, hp->h_length);
		return;
	}
	np = getnetbyname (s);
	if (np)
	{
		sin->sin_family = np->n_addrtype;
		sin->sin_addr = inet_makeaddr (np->n_net, INADDR_ANY);
		return;
	}
	sin->sin_family = AF_INET;
	val = inet_addr (s);
	if (val != -1)
	{
		sin->sin_addr.s_addr = val;
		return;
	}
	val = inet_network (s);
	if (val != -1)
	{
		sin->sin_addr = inet_makeaddr ((int)val, INADDR_ANY);
		return;
	}
	adios (NULLCP, "Unknown gateway address");
}

/* VARARGS2 */

static void  adios (what, fmt, a, b, c, d, e, f, g, h, i, j)
char   *what,
       *fmt,
       *a,
       *b,
       *c,
       *d,
       *e,
       *f,
       *g,
       *h,
       *i,
       *j;
{
    advise (what, LOG_ERR, fmt, a, b, c, d, e, f, g, h, i, j);

    _exit (1);
}


/* VARARGS3 */

void	advise (what, code, fmt, a, b, c, d, e, f, g, h, i, j)
char   *what,
       *fmt,
       *a,
       *b,
       *c,
       *d,
       *e,
       *f,
       *g,
       *h,
       *i,
       *j;
int	code;
{
    extern int errno;
    int	    eindex = errno;
    char    buffer[BUFSIZ];
    char    *sys_errname ();

    (void) sprintf (buffer, fmt, a, b, c, d, e, f, g, h, i, j);

    if (code < LOG_DEBUG) {
	if (what)
	    syslog (code, "%s %s: %s", buffer, what, sys_errname (eindex));
	else
	    syslog (code, "%s", buffer);
    }
    
    if (debug) {
	fprintf (stdout, "[%d] %s", code, buffer);
	if (what)
	    fprintf (stdout, " %s: %s", what, sys_errname (eindex));
	(void) fputc ('\n', stdout);
	(void) fflush (stdout);
    }

    errno = eindex;
}

char	*sys_errname (n)
int	n;
{
	static	char	buffer[64];
	extern int sys_nerr;
	extern	char	*sys_errlist[];

	if (0 < n && n < sys_nerr)
		return sys_errlist[n];
	(void) sprintf (buffer, "Error %d", n);
	return buffer;
}

close_conn (cp, fd, reason)
Connection	*cp;
int	fd;
char	*reason;
{
	advise (NULLCP, LOG_INFO, "Circuit closed %s (%s, %s open)",
		px25addr(&cp->x25remote), reason, ptime (cp -> contime));
	if ( fd != cp -> x25fd)
		adios (NULLCP, "Close of tunnel!");

	if (debug)
		(void) log_cause_and_diag (fd);
	(void) x25_close (fd);
	fd2con[fd] = NULLCON;
	readfds &= ~(1 << fd);
	x25fds &= ~(1 << fd);
	cp -> x25fd = NOTOK;
	if (cp -> flags & CON_OOD)
		x25open --;
	if ((cp -> flags & (CON_SLAVE|CON_MASTER)) == CON_SLAVE)
	{
		(void) close (cp -> tunfd);
		readfds &= ~(1 << cp -> tunfd);
		cp -> tunfd = NOTOK;
		fd2con[cp->tunfd] = NULLCON;
	}
	cp -> flags &= ~CON_OOD;
}

close_unused ()
{
	Connection	*cp;

	time_t	expired;

	expired = currenttime - timeout;
	
	for (cp = cons; cp -> device; cp ++) {
		if ( (cp -> flags & CON_OOD) &&
		    cp -> x25fd != NOTOK && cp -> lastuse <= expired)
			close_conn (cp, cp -> x25fd, "timeout");
	}
}

read_data (file)
char	*file;
{
	FILE	*fp;
	int	sd;
	char	buffer[BUFSIZ];
	Connection *cp = cons;
	char	*p, *q;

	if ((sd = socket (AF_X25, SOCK_STREAM, 0)) == NOTOK)
		adios ("failed", "X25 socket call");
	if (ioctl (sd, X25_RD_HOSTADR, &myx25addr))
		adios ("failed", "ioctl X25_RD_HOSTADR");
	(void) close (sd);
	if ((fp = fopen (file, "r")) == NULL)
		adios (file, "Can't open configuration file");
	while ((p = fgets (buffer, sizeof buffer, fp)) != NULL)
	{
		if (*p == '#' || *p == '\n')
			continue;
		cp -> device = getstr (&p);
		cp -> interface = getstr (&p);
		cp -> if_addr = getstr (&p);
		cp -> dst_addr = getstr (&p);
		q = getstr (&p);
		if ( *q == 's' || *q == 'S')
			cp -> flags = CON_SLAVE;
		else if (*q == 'b' || *q == 'B')
			cp -> flags = CON_SLAVE|CON_MASTER;
		else if (*q == 'M' || *q == 'm')
			cp -> flags = CON_MASTER;
		else	adios (NULLCP, "Unknown mode for dev %s - use s/m/b",
			       cp -> device);
		cp -> lastuse = 0;
		cp -> tunfd = cp -> x25fd = -1;
		getx25addr (&cp->x25local, &p);
		getx25addr (&cp->x25remote, &p);
#ifdef DEBUG
		printcp ("read in", cp);
#endif
		if (++cp >= &cons[MAXCONS])
			adios (NULLCP, "too many potential connections");
	}
	(void) fclose (fp);
}

open_conn (cp)
Connection	*cp;
{
	if ((cp -> x25fd = start_x25_client(&cp -> x25local)) == NOTOK) {
		advise ("failed", LOG_WARNING, "start_x25_client");
		return NOTOK;
	}
	if (join_x25_server (cp -> x25fd, &cp -> x25local, &cp -> x25remote)
								   == NOTOK) {
		advise ("failed", LOG_WARNING, "join_x25_server");
		(void) x25_close (cp -> x25fd);
		cp -> x25fd = NOTOK;
		return NOTOK;
	}
	advise (NULLCP, LOG_INFO, "joined to %s", px25addr(&cp->x25remote));
	set_x25opts(cp->x25fd);
	x25open ++;
	fd2con[cp->x25fd] = cp;
	cp -> flags |= CON_OOD;
	cp -> contime = cp -> lastuse = currenttime;
	readfds |= (1 << cp -> x25fd);
	return 0;
}

start_x25_client (cdb)
CONN_DB	*cdb;
{
	int	sd;
	int	pgrp;

	if((sd = socket (AF_X25, SOCK_STREAM, 0)) == NOTOK)
		return NOTOK;

	pgrp = getpid ();
	if (ioctl (sd, SIOCSPGRP, &pgrp) == NOTOK) {
		advise ("ioctl",LOG_WARNING, "can't set process group");
		(void) x25_close (sd);
		return NOTOK;
	}

	if (bind (sd, (struct sockaddr *)cdb, sizeof(CONN_DB)) == NOTOK)
		advise ("failed", LOG_WARNING, "bind on %s", px25addr(cdb));
	    
#ifdef X25_WR_FACILITY
	if (ioctl (sd, X25_WR_FACILITY, &facilities) == NOTOK) {
		advise ("failed", LOG_WARNING, "ioctl X25_WR_FACILITY");
		(void) x25_close (sd);
		return NOTOK;
	}
#endif
	return sd;
}

int join_x25_server (fd, local, remote)
register int fd;
register CONN_DB	*local;
register CONN_DB	*remote;
{
#ifdef SUNOS4
	/*
	 * under SunOS 4.0 we can set the calling address, and tell
	 * SunLink to use automatic link selection.
	 */
	CONN_ADR	local_adr;
	int		ret;
	local_adr.hostlen = local->hostlen;
	bcopy( (char *) local->host , (char *) local_adr.host,
		sizeof( local_adr.host));
	
	if( ioctl( fd, X25_WR_LOCAL_ADR, &local_adr))
		advise ("failed", LOG_WARNING, "ioctl SIOCSPGRP");
		
	remote->hostlen |= ANY_LINK;
	ret =  connect (fd, (struct sockaddr *)remote, sizeof (CONN_DB));
	remote->hostlen &= ~ ANY_LINK;
	return ret;
#else SUNOS4
	return connect (fd, (struct sockaddr *)remote, sizeof (CONN_DB));
#endif SUNOS4
}

char	*getstr (str)
char	**str;
{
	char	*p = *str;
	char	*q;
	char	buffer[128];

	while (isspace (*p)) p ++;
	if ( *p == '\0' || *p == '\n')
		adios (NULLCP, "end of line detected while looking for string");
	
	q = buffer;
	while (*p != '\n' && *p != '\0' && !isspace(*p))
		*q++ = *p ++;
	*q = 0;
	*str = p;
	return strdup (buffer);
}

char	*strdup(s)
char	*s;
{
	extern  char	*malloc ();
	extern  char	*strcpy ();
	auto	char	*p;

	if((p = malloc((unsigned)strlen(s) + 1)) == (char *)0)
		adios("malloc", "Out of memmory in strdup <%d bytes needed>");
	return strcpy(p, s);
}

getx25addr (cdb, p)
CONN_DB	*cdb;
char	**p;
{
	char	buffer[128];
	char	*q;
	char	*str = *p;

	while (isspace(*str)) str ++;

	if (*str == '\0' || *str == '\n')
		adios (NULLCP, "End of line while reading DTE address");

	q = buffer;
	if (*str == '-') {
		cdb -> hostlen = myx25addr.hostlen;
		bcopy ((char *)myx25addr.host, (char *)cdb->host,
		       sizeof (cdb->host));
		str ++;
}
	else {
		while (isdigit (*str))
		*q++ = *str ++;
		*q = '\0';
		cdb -> hostlen = char2bcd (buffer, cdb -> host);
	}
	cdb -> datalen = 0;
	while (isspace (*str)) str ++;
	if (*str == '\0' || *str == '\n')
		return;
	cdb -> datalen = getsstr((char *)cdb -> data, &str);
	*p = str;
}

int     char2bcd (s, d)
register char   *s;
register u_char *d;
{
	register int    c,
	i;

	for (i = 0; *s; i++) {
		if ((c = *s++) >= 'a' && c <= 'f')
			c -= 'a' + 0x0a;
		else
			if (c >= 'A' && c <= 'F')
				c -= 'A' + 0x0a;
			else
				if (c >= '0' && c <= '9')
					c -= '0';
				else
					c = 0;

		if (i & 1)
			*d++ |= c & 0xf;
		else
			*d = (c & 0xf) << 4;
	}

	return i;
}


int     bcd2char (s, d, len)
register    u_char *s;
register char  *d;
int     len;
{
	register int    i, g;

	for (i = 0; i < len; i++) {
		g = s[i >> 1];
		if ((i & 1) == 0)
			g >>= 4;
		g &= 0xf;

		if (g < 0x0a)
			*d++ = g + '0';
		else
			*d++ = g + 'a' - 0x0a;
	}

	*d = NULL;

	return len;
}

getsstr (to, from)
char	*to, **from;
{
	char	*p = to;
	char	*str = *from;
	char	n;

	if (*str == '-') {
		*p = 0xcc;
		str ++;
		*from = str;
		return 1;
	}

	while (isxdigit(*str) && isxdigit(*(str+1)))
	{
		n = a2h(*str++) << 4;
		n |= a2h(*str++);
		*p++ = n;
	}
	*from = str;
	return p - to;
}

a2h (c)
char	c;
{
	if (isdigit(c)) return c - '0';
	if (isupper(c)) return c - 'A' + 10;
	if (islower(c)) return c - 'a' + 10;
	return 0;
}

printcp(str, cp)
char	*str;
Connection *cp;
{

	if (!debug) return;
	fprintf (stdout,"Connection table entry %d - %s\n", cp - cons, str);
	fprintf (stdout,"\t%s %s %s %s 0x%x %d %d %d\n",
		cp -> device, cp->interface, cp -> if_addr, cp ->dst_addr,
		cp -> flags, cp -> lastuse, cp -> tunfd, cp -> x25fd);
	fprintf (stdout,"\t%s", px25addr (&cp-> x25local));
	fprintf (stdout," %s\n", px25addr (&cp-> x25remote));
}

char	*px25addr (cdb)
CONN_DB *cdb;
{
	static  char buffer[512];
	char	addr[128];
	char	cudf[128];

	(void) bcd2char(cdb->host, addr, (int)cdb->hostlen);
	(void) bcd2char(cdb->data, cudf, (int)cdb->datalen * 2);
	(void) sprintf (buffer, "[%d]%s [%d]%s", cdb -> hostlen, addr,
			cdb->datalen, cudf);
	return buffer;
}

int     start_x25_server (local, backlog, opt1, opt2)
CONN_DB	*local;
int     backlog,
	opt1,
	opt2;
{
	int     sd, pgrp;
	int	i;
	MASK_DATA_DB	mask;

	if ((sd = socket (AF_X25, SOCK_STREAM, 0)) == NOTOK)
		return NOTOK;	/* Can't get an X.25 socket */

	pgrp = getpid();
	if (ioctl(sd, SIOCSPGRP, &pgrp)) {
		advise ("failed", LOG_WARNING, "ioctl SIOCSPGRP");
		(void) x25_close (sd);
		return NOTOK;	/* Error can be found in errno */
	}

#ifdef X25_WR_MASK
	if (usemasks) {
		bzero ((char *)&mask, sizeof (mask));
		mask.masklen = local -> datalen;
		for (i = 0; i < mask.masklen &&
		     i << (sizeof mask.mask / sizeof mask.mask[0]); i++)
			mask.mask[i] = 0xff;

		if (ioctl (sd, X25_WR_MASK_DATA, &mask)) {
			advise ("failed", LOG_WARNING, "ioctl X25_WR_MASK_DATA");
			(void) x25_close (sd);
			return NOTOK;
		}
	}
#endif
	advise (NULLCP, LOG_INFO, "listening on %s", px25addr (local));
	if (bind (sd, (struct sockaddr *)local, sizeof(CONN_DB)) != NOTOK) {
		if (opt1)
			(void) setsockopt (sd, SOL_SOCKET, opt1, NULLCP, 0);
		if (opt2)
			(void) setsockopt (sd, SOL_SOCKET, opt2, NULLCP, 0);

		(void) listen (sd, backlog);
		return sd;
	}

	(void) x25_close (sd);
	return NOTOK;
}

int     join_x25_client (fd, remote)
int     fd;
CONN_DB *remote;
{
	int     len = sizeof(CONN_DB);

	return accept (fd, (struct sockaddr *) remote, &len);
}

envinit ()
{
	int	fd, i;

	if (debug)
		return;

	switch (fork()) {
	    case NOTOK:
		adios ("fork", "can't");

	    case 0:
		break;

	    default:
		exit (0);
	}
	for (i = getdtablesize (); i >= 0; i --)
		(void) close (i);
	if ((fd = open("/dev/null", 2)) != NOTOK) {
		(void) dup2(fd, 1);
		(void) dup2(fd, 2);
	}

#ifdef TIOCNOTTY
	if((fd = open("/dev/tty", 2)) != NOTOK) {
		(void) ioctl (fd, TIOCNOTTY, 0);
		(void) close (fd);
	}
#endif
#ifdef LOG_DAEMON
	(void) openlog (myname, LOG_PID,LOG_DAEMON);
#else
	(void) openlog (myname, LOG_PID);
#endif
}

x25cmp (cd1, cd2)
CONN_DB	*cd1, *cd2;
{
	int	len;
	char	buf1[16], buf2[16];

	(void) bcd2char(cd1->host, buf1, (int)cd1 -> hostlen);
	(void) bcd2char(cd2->host, buf2, (int)cd2 -> hostlen);

	len = cd1->hostlen < 12 ? cd1 -> hostlen : 12;
	if (strncmp (buf1, buf2, len) == 0)
		return 1;
	return 0;
}

char	*ptime (t)
time_t	t;
{
	static char buffer[BUFSIZ];
	char	*p;
	time_t	total;

	total = currenttime - t;
	p = buffer;
	*p = '\0';

	if (total <= 0) return "no time at all";
	if (total > 24 * 60 * 60) {
		sprintf (p, "%ld days", total / (24*60*60));
		total %= 24 * 60 * 60;
		p += strlen (p);
	}
	if (total > 60 * 60) {
		sprintf (p, " %ld hours", total / (60 * 60));
		total %= 60 * 60;
		p += strlen (p);
	}
	if (total > 60) {
		sprintf (p, " %ld minutes", total / 60);
		total %= 60;
		p += strlen (p);
	}
	if (total > 0) {
		sprintf (p, " %ld seconds", total);
	}
	return buffer;
}

/*
 * SunLink X.25 V.60beta has a problem with closing X.25
 * file descriptors. I do not know the exact circumstances,
 * but anyway, this routine tries to match the problem.
 */

void get_alarm() {
	return;
}

x25_close(fd)
{
	int last;
	struct sigvec old,new;

	new.sv_handler =  get_alarm;
	new.sv_flags = SV_INTERRUPT | SV_RESETHAND;
	new.sv_mask = sigmask(SIGALRM);
	sigvec(SIGALRM,&new,&old);

	last = alarm(30); /* > t3 */

	if(close(fd))
		advise ("failed", LOG_WARNING, "x25_close");

	sigvec(SIGALRM,&old,0);
	alarm(last);
}


/* handle SIGURG, coming from X.25 */

void oob_data()
{
	int fd,bits,okay,oob,error;
	char intr;
	char *type;

	/* check all x25 sockets that we are interested in */
	okay = 0;
	fd = 0;
	bits = x25fds;
	while( bits) {
		if( bits & 1) {
			if( error = ioctl(fd,X25_OOB_TYPE,&oob))
				advise("failed", LOG_WARNING, "SIGURG: ioctl X25_OOB_TYPE on fd %d", fd);
			if( oob) {
				okay++;
				switch(oob) {
				case VC_RESET : advise(NULLCP,LOG_WARNING,"SIGURG: VC_RESET on fd %d",fd);
				break;
				case N_RESETS : advise(NULLCP,LOG_WARNING,"SIGURG: N_RESETS on fd %d",fd);
				break;
				case N_CLEARS : advise(NULLCP,LOG_WARNING,"SIGURG: N_CLEARS on fd %d",fd);
				break;
				case MSG_TOO_LONG : advise(NULLCP,LOG_WARNING,"SIGURG: MSG_TOO_LONG on fd %d",fd);
				break;
				case INT_DATA :
					/* read interrupt type */
					error = recv(fd, &intr,1,MSG_OOB);
					if(error)
						advise("failed", LOG_WARNING, "SIGURG: recv intr packet");
					advise(NULLCP,LOG_WARNING,"SIGURG: INT_DATA (%d), on fd %d",intr, fd);
					break;
				default	:
					advise(NULLCP,LOG_WARNING,"SIGURG: unknown oob data: %d",oob);
				}

			}
		}
		bits >>= 1;
		fd++;
	}
	if(!okay)
		advise(NULLCP,LOG_WARNING,"SIGURG: found no socket with OOB_DATA");

}

/* change high water mark for x25 socket */

void set_x25opts(fd)
int fd;
{
	SO_HIWAT_DB hiwater;
	int error,flags;

	error = ioctl(fd, X25_RD_SBHIWAT, &hiwater);
	if(error) {
		advise("failed",LOG_WARNING,"ioctl X25_RD_SBHIWAT");
		return;
	}
	advise(NULLCP,LOG_INFO,"read hiwater on fd %d: %d, %d",
		fd,hiwater.sendhiwat,hiwater.recvhiwat);
	hiwater.sendhiwat = 32767;
	hiwater.recvhiwat = 32767;
	error = ioctl(fd, X25_WR_SBHIWAT, &hiwater);
	if(error) {
		advise("failed",LOG_WARNING,"ioctl X25_WR_SBHIWAT");
		return;
	}
	advise(NULLCP,LOG_INFO,"set hiwater on fd %d: %d, %d",
		fd,hiwater.sendhiwat,hiwater.recvhiwat);

	/*

	flags = fcntl (fd,F_GETFL,0);
	if( flags == NOTOK) {
		advise("failed",LOG_WARNING,"fcntl F_GETFL");
		return;
	}
	if( fcntl (fd,F_SETFL, flags | FNDELAY) == NOTOK) {
		advise("failed",LOG_WARNING,"fcntl F_SETFL");
	}

	*/

	return;
}
